from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
INSUFFICIENT_FUNDS: StatusCode
INVALID: StatusCode
NOT_FOUND: StatusCode
OK: StatusCode
SUBSCRIPTION_REQUIRED: StatusCode
UNAUTHORIZED: StatusCode
UNKNOWN: StatusCode

class Delta(_message.Message):
    __slots__ = ["add_dir", "add_path", "remove_path"]
    ADD_DIR_FIELD_NUMBER: _ClassVar[int]
    ADD_PATH_FIELD_NUMBER: _ClassVar[int]
    REMOVE_PATH_FIELD_NUMBER: _ClassVar[int]
    add_dir: _containers.RepeatedScalarFieldContainer[str]
    add_path: _containers.RepeatedScalarFieldContainer[str]
    remove_path: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, add_path: _Optional[_Iterable[str]] = ..., add_dir: _Optional[_Iterable[str]] = ..., remove_path: _Optional[_Iterable[str]] = ...) -> None: ...

class DockerImagePullProgress(_message.Message):
    __slots__ = ["current", "id", "status", "total"]
    CURRENT_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    current: int
    id: str
    status: str
    total: int
    def __init__(self, id: _Optional[str] = ..., status: _Optional[str] = ..., total: _Optional[int] = ..., current: _Optional[int] = ...) -> None: ...

class Environment(_message.Message):
    __slots__ = ["alloc", "python", "sh"]
    ALLOC_FIELD_NUMBER: _ClassVar[int]
    PYTHON_FIELD_NUMBER: _ClassVar[int]
    SH_FIELD_NUMBER: _ClassVar[int]
    alloc: Remote
    python: Python
    sh: str
    def __init__(self, alloc: _Optional[_Union[Remote, _Mapping]] = ..., python: _Optional[_Union[Python, _Mapping]] = ..., sh: _Optional[str] = ...) -> None: ...

class ExecRequest(_message.Message):
    __slots__ = ["argv", "cwd", "rsync_source", "stdin", "workspace_id"]
    ARGV_FIELD_NUMBER: _ClassVar[int]
    CWD_FIELD_NUMBER: _ClassVar[int]
    RSYNC_SOURCE_FIELD_NUMBER: _ClassVar[int]
    STDIN_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    argv: _containers.RepeatedScalarFieldContainer[str]
    cwd: str
    rsync_source: RsyncSource
    stdin: bytes
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ..., argv: _Optional[_Iterable[str]] = ..., rsync_source: _Optional[_Union[RsyncSource, _Mapping]] = ..., cwd: _Optional[str] = ..., stdin: _Optional[bytes] = ...) -> None: ...

class ExecResponse(_message.Message):
    __slots__ = ["execution_id", "output_files", "result", "stderr", "stdout"]
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FILES_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STDERR_FIELD_NUMBER: _ClassVar[int]
    STDOUT_FIELD_NUMBER: _ClassVar[int]
    execution_id: str
    output_files: _containers.RepeatedScalarFieldContainer[str]
    result: Result
    stderr: bytes
    stdout: bytes
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., execution_id: _Optional[str] = ..., stdout: _Optional[bytes] = ..., stderr: _Optional[bytes] = ..., output_files: _Optional[_Iterable[str]] = ...) -> None: ...

class GetUserResponse(_message.Message):
    __slots__ = ["result", "username"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    result: Result
    username: str
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., username: _Optional[str] = ...) -> None: ...

class Hardware(_message.Message):
    __slots__ = ["disks", "processor"]
    class Disks(_message.Message):
        __slots__ = ["other", "rxroot_size"]
        class DiskSpec(_message.Message):
            __slots__ = ["mount_point", "size", "source"]
            MOUNT_POINT_FIELD_NUMBER: _ClassVar[int]
            SIZE_FIELD_NUMBER: _ClassVar[int]
            SOURCE_FIELD_NUMBER: _ClassVar[int]
            mount_point: str
            size: str
            source: str
            def __init__(self, size: _Optional[str] = ..., mount_point: _Optional[str] = ..., source: _Optional[str] = ...) -> None: ...
        OTHER_FIELD_NUMBER: _ClassVar[int]
        RXROOT_SIZE_FIELD_NUMBER: _ClassVar[int]
        other: _containers.RepeatedCompositeFieldContainer[Hardware.Disks.DiskSpec]
        rxroot_size: str
        def __init__(self, rxroot_size: _Optional[str] = ..., other: _Optional[_Iterable[_Union[Hardware.Disks.DiskSpec, _Mapping]]] = ...) -> None: ...
    DISKS_FIELD_NUMBER: _ClassVar[int]
    PROCESSOR_FIELD_NUMBER: _ClassVar[int]
    disks: Hardware.Disks
    processor: str
    def __init__(self, processor: _Optional[str] = ..., disks: _Optional[_Union[Hardware.Disks, _Mapping]] = ...) -> None: ...

class InitRequest(_message.Message):
    __slots__ = ["project_name", "rsync_source", "target_env"]
    PROJECT_NAME_FIELD_NUMBER: _ClassVar[int]
    RSYNC_SOURCE_FIELD_NUMBER: _ClassVar[int]
    TARGET_ENV_FIELD_NUMBER: _ClassVar[int]
    project_name: str
    rsync_source: RsyncSource
    target_env: Environment
    def __init__(self, project_name: _Optional[str] = ..., rsync_source: _Optional[_Union[RsyncSource, _Mapping]] = ..., target_env: _Optional[_Union[Environment, _Mapping]] = ...) -> None: ...

class InitResponse(_message.Message):
    __slots__ = ["payment", "result", "rsync_dest", "worker_addr", "workspace_id"]
    class PaymentRequired(_message.Message):
        __slots__ = ["needs_metered", "needs_subscription"]
        NEEDS_METERED_FIELD_NUMBER: _ClassVar[int]
        NEEDS_SUBSCRIPTION_FIELD_NUMBER: _ClassVar[int]
        needs_metered: float
        needs_subscription: bool
        def __init__(self, needs_subscription: bool = ..., needs_metered: _Optional[float] = ...) -> None: ...
    PAYMENT_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    RSYNC_DEST_FIELD_NUMBER: _ClassVar[int]
    WORKER_ADDR_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    payment: InitResponse.PaymentRequired
    result: Result
    rsync_dest: RsyncDestination
    worker_addr: str
    workspace_id: str
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., rsync_dest: _Optional[_Union[RsyncDestination, _Mapping]] = ..., worker_addr: _Optional[str] = ..., workspace_id: _Optional[str] = ..., payment: _Optional[_Union[InitResponse.PaymentRequired, _Mapping]] = ...) -> None: ...

class InstallDepsRequest(_message.Message):
    __slots__ = ["workspace_id"]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ...) -> None: ...

class InstallDepsResponse(_message.Message):
    __slots__ = ["result", "stdout"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STDOUT_FIELD_NUMBER: _ClassVar[int]
    result: Result
    stdout: bytes
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., stdout: _Optional[bytes] = ...) -> None: ...

class KillRequest(_message.Message):
    __slots__ = ["execution_id", "workspace_id"]
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    execution_id: str
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ..., execution_id: _Optional[str] = ...) -> None: ...

class Python(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class Remote(_message.Message):
    __slots__ = ["hardware", "image", "ports"]
    class Image(_message.Message):
        __slots__ = ["docker", "tag"]
        DOCKER_FIELD_NUMBER: _ClassVar[int]
        TAG_FIELD_NUMBER: _ClassVar[int]
        docker: str
        tag: str
        def __init__(self, docker: _Optional[str] = ..., tag: _Optional[str] = ...) -> None: ...
    HARDWARE_FIELD_NUMBER: _ClassVar[int]
    IMAGE_FIELD_NUMBER: _ClassVar[int]
    PORTS_FIELD_NUMBER: _ClassVar[int]
    hardware: Hardware
    image: Remote.Image
    ports: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, image: _Optional[_Union[Remote.Image, _Mapping]] = ..., hardware: _Optional[_Union[Hardware, _Mapping]] = ..., ports: _Optional[_Iterable[int]] = ...) -> None: ...

class Result(_message.Message):
    __slots__ = ["code", "message"]
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: StatusCode
    message: str
    def __init__(self, code: _Optional[_Union[StatusCode, str]] = ..., message: _Optional[str] = ...) -> None: ...

class RsyncDestination(_message.Message):
    __slots__ = ["daemon_module"]
    DAEMON_MODULE_FIELD_NUMBER: _ClassVar[int]
    daemon_module: str
    def __init__(self, daemon_module: _Optional[str] = ...) -> None: ...

class RsyncSource(_message.Message):
    __slots__ = ["directory", "machine_id", "public_key"]
    DIRECTORY_FIELD_NUMBER: _ClassVar[int]
    MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    PUBLIC_KEY_FIELD_NUMBER: _ClassVar[int]
    directory: str
    machine_id: int
    public_key: bytes
    def __init__(self, machine_id: _Optional[int] = ..., directory: _Optional[str] = ..., public_key: _Optional[bytes] = ...) -> None: ...

class SetUsernameRequest(_message.Message):
    __slots__ = ["username"]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    username: str
    def __init__(self, username: _Optional[str] = ...) -> None: ...

class SetUsernameResponse(_message.Message):
    __slots__ = ["result"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: Result
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ...) -> None: ...

class StopRequest(_message.Message):
    __slots__ = ["workspace_id"]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ...) -> None: ...

class StopResponse(_message.Message):
    __slots__ = ["result"]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: Result
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ...) -> None: ...

class WorkerInitRequest(_message.Message):
    __slots__ = ["workspace_id"]
    WORKSPACE_ID_FIELD_NUMBER: _ClassVar[int]
    workspace_id: str
    def __init__(self, workspace_id: _Optional[str] = ...) -> None: ...

class WorkerInitResponse(_message.Message):
    __slots__ = ["pull_progress", "result"]
    PULL_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    pull_progress: DockerImagePullProgress
    result: Result
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., pull_progress: _Optional[_Union[DockerImagePullProgress, _Mapping]] = ...) -> None: ...

class StatusCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



"""Checks that packaging worked correctly."""
import os
from packaging import version
import re
import tarfile
from typing import Tuple
import unittest

from absl import flags
import toml

from rx.client.configuration import local

FLAGS = flags.FLAGS

IGNORE = [
  '__pycache__',
  '^.git',
  '^.pylintrc',
  '^.pytest_cache',
  '^.rx',
  '^.rxignore',
  '^.venv',
  '^.vscode',
  '^build',
  '^dist',
  '^integration_test.sh$',
  '^release.sh',
  '^requirements.txt',
  '^requirements_dev.txt',
  '^run_tests.py$',
  '^rx-out',
  '\\btest_.*.py$',
  '\\b__init__.py$',
  '.proto$',
]

REMOVE_REQS = [
  'certifi',
  'charset-normalizer',
  'idna',
  'urllib3',
]


class PackagingTests(unittest.TestCase):
  def test_tgz_contains_all_dirs(self):
    tgz_version, tgz_path = _get_latest_tgz()
    tarf = tarfile.open(tgz_path, 'r:gz')
    tar_ent = tarf.getnames()
    for path, _, files in os.walk('.'):
      for name in files:
        target = os.path.normpath(os.path.join(path, name))
        if os.path.isdir(target):
          # tar file doesn't contain plain dirs.
          continue
        if _ignorable(target):
          continue
        # Everything in the package is prefixed with 'run_rx-version/'
        pkg_path = os.path.join(f'run_rx-{tgz_version}', target)
        self.assertIn(
          pkg_path, tar_ent, f'Missing file {pkg_path} in {tgz_path}')


class TomlTests(unittest.TestCase):

  def setUp(self) -> None:
    super().setUp()
    with open('pyproject.toml', mode='rt', encoding='utf-8') as fh:
      self._pyproject = toml.load(fh)

  def test_version_matches_in_code_and_config(self):
    self.assertEqual(
      local.VERSION, self._pyproject['tool']['poetry']['version'])

  def test_requirements_all_there(self):
    got = list(self._pyproject['tool']['poetry']['dependencies'].keys())
    got.remove('python')

    reqs = open('requirements.txt', 'r')
    expected = [x.split('==')[0] for x in reqs.readlines()]
    for rr in REMOVE_REQS:
      expected.remove(rr)
    self.assertEqual(got, expected)


def _get_latest_tgz() -> Tuple[version.Version, str]:
  dist = os.listdir('dist')
  max_version = version.parse('0.0.1')
  max_tgz = None
  # Format of each file: run_rx-x.y.z.tar.gz
  for f in dist:
    if not f.endswith('tar.gz'):
      # Wheel file.
      continue
    v = version.parse(f[7:-7])
    if v > max_version:
      max_version = v
      max_tgz = f
  assert max_tgz is not None, 'No .tar.gzs in dist/'
  return max_version, os.path.join('dist', max_tgz)


def _ignorable(target: str):
  for x in IGNORE:
    if re.search(x, target) is not None:
      return True
  return False



import unittest

from rx.client import user


class UserTests(unittest.TestCase):

  def test_username_too_short(self):
    self.assertFalse(user.username_validator(None))
    self.assertFalse(user.username_validator(''))
    self.assertFalse(user.username_validator('x'))

  def test_username_too_long(self):
    self.assertFalse(user.username_validator('a012345678901234'))

  def test_username_invalid(self):
    self.assertFalse(user.username_validator('123abc'))
    self.assertFalse(user.username_validator('x.y.com'))

  def test_valid_username(self):
    # Min length.
    self.assertTrue(user.username_validator('abc'))
    self.assertTrue(user.username_validator('abc123'))
    self.assertTrue(user.username_validator('abc_123'))
    # Max length.
    self.assertTrue(user.username_validator('a01234567890123'))



import pathlib
import sys
import unittest
from unittest import mock

from absl import flags
from absl.testing import absltest

from rx.client.worker import rsync
from rx.client.configuration import local

FLAGS = flags.FLAGS


class MockLocalConfig(local.LocalConfig):
  def __init__(self, config):
    self._cwd = '/path/to/proj'
    self._config = config


class RsyncTests(unittest.TestCase):

  def setUp(self) -> None:
    super().setUp()
    FLAGS(sys.argv)
    self._cwd = pathlib.Path('/path/to/proj')
    self._local_cfg = MockLocalConfig({'rsync_path': '/usr/bin/rsync'})
    self._remote_cfg = {
      'daemon_module': 'f1d1df3b-e046-4e88-822e-72596e5020c5',
      'worker_addr': 'abc123.trex.run-rx.com',
      'workspace_id': 'f1d1df3b-e046-4e88-822e-72596e5020c5',
    }

  def test_from_remote(self):
    client = rsync.RsyncClient(self._local_cfg, self._remote_cfg)
    rsync._run_rsync = mock.MagicMock(return_value=0)

    outdir = pathlib.Path(absltest.TEST_TMPDIR.value)
    outdir.mkdir(parents=True, exist_ok=True)
    got = client.from_remote('rx-out', outdir)

    self.assertEqual(got, 0)
    rsync._run_rsync.assert_called_once_with([
      '/usr/bin/rsync',
      '--archive',
      '--compress',
      '--delete',
      '--quiet',
      '--exclude-from=/path/to/proj/.rxignore',
      'abc123.trex.run-rx.com::f1d1df3b-e046-4e88-822e-72596e5020c5/rx-out/',
      str(outdir),
    ])

  def test_to_remote(self):
    client = rsync.RsyncClient(self._local_cfg, self._remote_cfg)
    rsync._run_rsync = mock.MagicMock(return_value=0)

    got = client.to_remote()

    self.assertEqual(got, 0)
    rsync._run_rsync.assert_called_once_with([
      '/usr/bin/rsync',
      '--archive',
      '--compress',
      '--delete',
      '--inplace',
      '--exclude-from=/path/to/proj/.rxignore',
      '/path/to/proj/',
      'abc123.trex.run-rx.com::f1d1df3b-e046-4e88-822e-72596e5020c5',
    ])


if __name__ == '__main__':
  absltest.main()


import pathlib

from rx.client.configuration import config_base
from rx.client.configuration import local

class Command:
  def __init__(self) -> None:
    if config_base.RX_ROOT.value:
      self._rxroot = pathlib.Path(config_base.RX_ROOT.value)
    else:
      self._rxroot = local.find_rxroot(pathlib.Path.cwd())



import fcntl
import os
import sys
import selectors
import termios
import threading
import tty
from types import TracebackType
from typing import Any, Optional, Tuple, Type

from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

STDOUT = 1
STDERR = 2


class Executor:

  def __init__(
      self,
      stub: rx_pb2_grpc.ExecutionServiceStub,
      req: rx_pb2.ExecRequest) -> None:
    self._stub = stub
    self._request = req

  def run(
      self, metadata: Tuple[Tuple[str, Any], ...]
  ) -> Optional[rx_pb2.ExecResponse]:
    response = None
    with StdinIterator(self._request) as req_it:
      for response in self._stub.Exec(req_it, metadata=metadata):
        self.write(response)
    return response

  def write(self, resp: rx_pb2.ExecResponse) -> None:
    if resp.stdout:
      sys.stdout.buffer.write(resp.stdout)
      sys.stdout.flush()
    if resp.stderr:
      sys.stderr.buffer.write(resp.stderr)
      sys.stderr.flush()


class StdinIterator:

  def __init__(self, request: rx_pb2.ExecRequest) -> None:
    self._initial_request = request
    self._running = True
    self._stdin_buffer = [request]
    self._stdin_avail = threading.Condition()
    self._original_attrs = termios.tcgetattr(sys.stdin)

  def __enter__(self) -> 'StdinIterator':
    orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl | os.O_NONBLOCK)
    th = threading.Thread(target=self.loop, daemon=True)
    th.start()
    return self

  def __exit__(self, exctype: Optional[Type[BaseException]],
             excinst: Optional[BaseException],
             exctb: Optional[TracebackType]):
    termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self._original_attrs)
    self._running = False
    return False

  def __next__(self) -> rx_pb2.ExecRequest:
    # This is probably unnecessary, nothing should call next once exec is done.
    if not self._running:
      raise StopIteration()
    with self._stdin_avail:
      while not self._stdin_buffer:
        if not self._running:
          raise StopIteration()
        self._stdin_avail.wait()
      return self._stdin_buffer.pop(0)

  def loop(self):
    # Prevent echo of characters.
    tty.setcbreak(sys.stdin)

    sel = selectors.DefaultSelector()
    sel.register(sys.stdin, selectors.EVENT_READ, self.got_stdin)

    while self._running:
      events = sel.select(timeout=0)
      for key, mask in events:
        del mask
        key.data()

  def got_stdin(self):
    stdin_str = sys.stdin.read()
    if stdin_str:
      with self._stdin_avail:
        self._stdin_buffer += [
          rx_pb2.ExecRequest(stdin=stdin_str.encode('utf-8'))]
        self._stdin_avail.notify()




import os
import pathlib
import sys
from typing import List, Tuple, cast

from absl import flags
from absl import logging
import grpc
import tqdm

from rx.client import login
from rx.client import output_handler
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.client.worker import executor
from rx.client.worker import rsync
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

SIGINT_CODE = 130

# TODO: set up UI for looking at previous executions.
_BASE_URI = 'https://la-brea.run-rx.com'
_NOT_REACHABLE = 1

_CWD = flags.DEFINE_string('cwd', None, 'Directory to run the command from')

class Client:
  """Handle contacting the remote server."""

  def __init__(
      self,
      channel: grpc.Channel,
      local_cfg: local.LocalConfig,
      auth_metadata: Tuple[Tuple[str, str]]):
    self._local_cfg = local_cfg
    self._remote_cfg = remote.Remote(local_cfg.cwd)
    self._rsync = rsync.RsyncClient(local_cfg, self._remote_cfg)
    self._stub = rx_pb2_grpc.ExecutionServiceStub(channel)
    self._metadata = local.get_grpc_metadata() + auth_metadata
    self._current_execution_id = None

  def init(self):
    # Get the container downloaded/running.
    req = rx_pb2.WorkerInitRequest(workspace_id=self._remote_cfg.workspace_id)
    progress_bars = {}
    try:
      for resp in self._stub.Init(req, metadata=self._metadata):
        if resp.result.code != 0:
          raise InitError(
            f'Error initializing worker {self._remote_cfg.worker_addr}: '
            f'{resp.result.message}')
        if resp.pull_progress:
          pp = resp.pull_progress
          if pp.id not in progress_bars:
            # First item must have a total.
            if pp.total == 0:
              continue
            progress_bars[pp.id] = ProgressBar(pp)
          progress_bars[pp.id].update(pp)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise InitError(
        f'Error initializing worker {self._remote_cfg.worker_addr}: '
        f'{e.details()}')
    for p in progress_bars.values():
      p.close()

    # Sync sources.
    result = self._rsync.to_remote()
    if result != 0:
      raise UnreachableError(self._remote_cfg.worker_addr, result)

    # Install deps.
    self._install_deps()

  def exec(self, argv: List[str]) -> int:

    cmd_str = ' '.join(argv)
    logging.info(f'Running `{cmd_str}` on {self._remote_cfg.worker_addr}')

    result = self._rsync.to_remote()
    if result != 0:
      raise UnreachableError(self._remote_cfg.worker_addr, result)

    rxroot = os.path.abspath(self._local_cfg.cwd)
    cwd = os.path.abspath(pathlib.Path.cwd())
    if is_subdir(parent=rxroot, child=cwd):
      cwd = str(pathlib.Path.cwd().relative_to(self._local_cfg.cwd))
    else:
      cwd = _CWD.value

    out_handler = output_handler.OutputHandler(self._local_cfg.cwd)

    request = rx_pb2.ExecRequest(
      workspace_id=self._remote_cfg.workspace_id, argv=argv, cwd=cwd)
    runner = executor.Executor(self._stub, request)
    try:
      response = runner.run(self._metadata)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      sys.stderr.write(f'Error contacting {self._remote_cfg.worker_addr}: {e.details()}\n')
      return _NOT_REACHABLE
    except KeyboardInterrupt:
      self.maybe_kill()
      return SIGINT_CODE
    except RuntimeError as e:
      # Bug in grpc, probably.
      msg = str(e)
      if msg == 'cannot release un-acquired lock':
        self.maybe_kill()
        return SIGINT_CODE
      else:
        sys.stderr.write(msg)
        return -1

    out_handler.write_outputs(self._rsync)

    if (response is not None and response.HasField('result') and
        response.result.code != 0):
      sys.stderr.write(f'{response.result.message}\n')
      return response.result.code
    return 0

  def maybe_kill(self):
    """Kill the process, if it exists."""
    if self._current_execution_id is None:
      logging.error('No ID to kill.')
      return

    req = rx_pb2.KillRequest(
      workspace_id=self._remote_cfg.workspace_id,
      execution_id=self._current_execution_id,
    )
    self._stub.Kill(req, metadata=self._metadata)

  def _install_deps(self):
    req = rx_pb2.InstallDepsRequest(workspace_id=self._remote_cfg.workspace_id)
    resp = None
    try:
      for resp in self._stub.InstallDeps(
        req, metadata=self._metadata, timeout=(10 * 60)):
        if resp.stdout:
          sys.stdout.buffer.write(resp.stdout)
          sys.stdout.buffer.flush()
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise InitError(e.details(), -1)
    if resp and resp.HasField('result') and resp.result.code:
      raise InitError(resp.result.message, resp.result.code)


def create_authed_client(ch: grpc.Channel, local_cfg: local.LocalConfig):
  lm = login.LoginManager(local_cfg.cwd)
  lm.login()
  return Client(ch, local_cfg, lm.grpc_metadata)


def is_subdir(*, parent: str, child: str) -> bool:
  return os.path.commonpath([parent]) == os.path.commonpath([parent, child])


class ProgressBar:

  def __init__(self, pp: rx_pb2.DockerImagePullProgress) -> None:
    assert pp.total > 0
    self._bar = tqdm.tqdm(
      desc=f'{pp.status} layer {pp.id}',
      total=pp.total,
      unit=' bytes')
    self._is_done = False

  def close(self):
    self._bar.close()

  def update(self, pp: rx_pb2.DockerImagePullProgress):
    if pp.total == 0:
      self._bar.set_description(f'{pp.status} layer {pp.id}')
      return
    if self._is_done:
      return

    # Note: this jumps back to 0 for download -> extract, which is good?
    self._bar.update(pp.current - self._bar.n)
    if self._bar.n == pp.total:
      self._is_done = True


# TODO: this needs to take code.
class InitError(RuntimeError):
  pass

class UnreachableError(RuntimeError):
  def __init__(self, worker_addr: str, code: int, *args: object) -> None:
    super().__init__(*args)
    self.worker = worker_addr
    self.code = code



"""Text menus.

Each helper function has the same general structure:

1. Setup the prompt.
2. Convert the response to the expected type.
2. Validate the response. If validation fails, go back to 1.
"""

import pathlib
from typing import Any, Callable, List, Optional

from absl import flags

_QUIET = flags.DEFINE_bool(
    'quiet', False, 'Don\'t prompt for user confirmation, go with the default '
    'option automatically')


def bool_prompt(prompt: str, default_opt: str = '') -> bool:
  assert default_opt in ['', 'y', 'n']
  if _QUIET.value and default_opt:
    return default_opt == 'y'
  y = 'Y' if default_opt == 'y' else 'y'
  n = 'N' if default_opt == 'n' else 'n'
  response = input(f'{prompt} ({y}/{n}): ')
  if not response:
    response = default_opt
  response = response.lower()
  if response == 'y':
    return True
  elif response == 'n':
    return False
  print('Please choose "y" or "n".')
  return bool_prompt(prompt)


def numbered_options_prompt(
    prompt: str, options: List[Any], default: int = 0) -> int:
  """Prompt with a numbered list of options."""
  if _QUIET.value:
    return default

  # Setup
  assert options
  print(prompt)
  for i, p in options:
    print(f'{i}: {p}')
  response = input(f'[{default}]: ')

  # Convert.
  if not response:
    return 0
  response = int(response)

  # Validation.
  if response >= len(options):
    print(f'Must be between 0 and {len(options) - 1}')
    return numbered_options_prompt(prompt, options, default)
  return response


def path_prompt(
  prompt: str,
  default_path: Optional[pathlib.Path],
  validation: Callable[[Optional[pathlib.Path]], bool] = lambda x: True
) -> pathlib.Path:
  """Prompt for filesystem path."""
  if _QUIET.value and default_path:
    return default_path

  # Setup.
  if default_path:
    response = input(f'{prompt}\n[{default_path}]: ')
  else:
    response = input(f'{prompt}: ')

  # Convert.
  if not response and default_path:
    response = default_path
  else:
    response = pathlib.Path(response)

  # Validation.
  if not validation(response):
    return path_prompt(prompt, default_path, validation)
  return response


def string_prompt(
  prompt: str,
  default_str: Optional[str] = None,
  validation:  Callable[[Optional[str]], bool] = lambda x: True
) -> str:
  if _QUIET.value and default_str:
    return default_str

  # Setup.
  if default_str:
    response = input(f'{prompt}\n[{default_str}]: ')
  else:
    response = input(f'{prompt}: ')

  # Convert.
  if not response and default_str:
    response = default_str

  # Validation.
  if not validation(response):
    return string_prompt(prompt, default_str, validation)
  return response



import math
import time
import unittest

from rx.client import login

class LoginTests(unittest.TestCase):

  def test_expired_ts(self):
    past = math.floor(time.time()) - 10
    self.assertTrue(login.is_expired({'exp': past}))

  def test_not_expired(self):
    future = math.floor(time.time()) + 1000
    self.assertFalse(login.is_expired({'exp': future}))




import json
import pathlib
import shutil
import tempfile
from typing import Any, Dict
import unittest

from absl.testing import absltest
from absl.testing import flagsaver

from rx.proto import rx_pb2
from rx.client.configuration import local


class LocalTests(unittest.TestCase):

  def setUp(self) -> None:
    super().setUp()
    self._rxroot = pathlib.Path(tempfile.mkdtemp())
    local._install_local_files(self._rxroot)

  def tearDown(self) -> None:
    super().tearDown()
    shutil.rmtree(self._rxroot)

  def test_create_default_local_config(self):
    cfg = local.create_local_config(self._rxroot)

    self.assertEqual(cfg['project_name'], self._rxroot.name)
    self.assertEqual(cfg['remote'], '.rx/remotes/default')

  @flagsaver.flagsaver(remote='test')
  def test_create_target_env(self):
    remote_config = {
      "image": {
        "docker": "abc123"
      },
      "hardware": {
        "processor": "xpu"
      }
    }
    self._create_remote(remote_config)
    cfg = local.create_local_config(self._rxroot)

    got = cfg.get_target_env()

    expected = rx_pb2.Environment(
      alloc=rx_pb2.Remote(
        image=rx_pb2.Remote.Image(docker='abc123'),
        hardware=rx_pb2.Hardware(processor='xpu'),
      ),
      sh='/bin/bash',
    )
    self.assertEqual(got, expected)

  @flagsaver.flagsaver(remote='no-hardware')
  def test_no_hardware(self):
    remote_config = {
      "image": {
        "docker": "python:1.2.3"
      }
    }
    self._create_remote(remote_config, 'no-hardware')
    cfg = local.create_local_config(self._rxroot)

    got = cfg.get_target_env()

    expected = rx_pb2.Environment(
      alloc=rx_pb2.Remote(
        image=rx_pb2.Remote.Image(docker='python:1.2.3'),
      ),
      sh='/bin/bash',
    )
    self.assertEqual(got, expected)

  def _create_remote(self, cfg: Dict[str, Any], remote_name: str = 'test'):
    remote_config_file = self._rxroot / remote_name
    with remote_config_file.open(mode='wt', encoding='utf-8') as fh:
      json.dump(cfg, fh)


if __name__ == '__main__':
  absltest.main()



import io
import pathlib
import shutil
import sys
import unittest
from unittest import mock

from absl.testing import absltest
from absl.testing import flagsaver

from rx.client.commands import init
from rx.client.configuration import local


class InitTests(unittest.TestCase):

  def setUp(self) -> None:
    super().setUp()
    self.maxDiff = None
    shutil.rmtree(absltest.get_default_test_tmpdir())
    pathlib.Path(absltest.get_default_test_tmpdir()).mkdir()

  def test_rxroot_flag(self):
    tmpdir = absltest.get_default_test_tmpdir()
    rxroot = pathlib.Path(tmpdir) / 'foo'
    rxroot.mkdir(exist_ok=True)

    with flagsaver.flagsaver(rxroot=str(rxroot)):
      got = init.InitCommand()

    self.assertEqual(got.rxroot, rxroot)

  def test_no_rxroot(self):
    got = init.InitCommand()

    self.assertEqual(got.rxroot, pathlib.Path.cwd())

  def test_existing_rxroot_message(self):
    tmpdir = absltest.get_default_test_tmpdir()
    rxroot = pathlib.Path(tmpdir) / 'foo'
    rxroot.mkdir(exist_ok=True)
    local.create_local_config(rxroot)

    with flagsaver.flagsaver(rxroot=str(rxroot)), \
        mock.patch.object(sys, 'stdout', new=io.StringIO()) as mock_stdout, \
        mock.patch.object(sys, 'stdin', io.StringIO('y\ny')):
      cmd = init.InitCommand()
      cmd._show_init_message()

    expected = f"""Looks like you already have an rx workspace. Would you like to stop that one
and start a new one? (Y/n): Got it. To re-init this workspace, rx will:

1. Shut down your existing virtual machine.
2. Create an account for you with rx (or log you into your existing account).
3. Set up a virtual machine on the cloud (on AWS).
4. Copy over the files in this directory ({rxroot}) to your
   virtual machine.

Would you like to proceed with logging in/creating an rx account? (Y/n): """
    self.assertEqual(expected, mock_stdout.getvalue())

  def test_new_rxroot_message(self):
    tmpdir = absltest.get_default_test_tmpdir()
    rxroot = pathlib.Path(tmpdir) / 'foo'
    rxroot.mkdir(exist_ok=True)

    with flagsaver.flagsaver(rxroot=str(rxroot)), \
        mock.patch.object(sys, 'stdout', new=io.StringIO()) as mock_stdout, \
        mock.patch.object(sys, 'stdin', io.StringIO('y')):
      cmd = init.InitCommand()
      cmd._show_init_message()

    expected = f"""To set up rx, this command will:

1. Create an account for you with rx (or log you into your existing account).
2. Set up a virtual machine on the cloud (on AWS).
3. Copy over the files in this directory ({rxroot}) to your
   virtual machine.

Would you like to proceed with logging in/creating an rx account? (Y/n): """
    self.assertEqual(expected, mock_stdout.getvalue())


if __name__ == '__main__':
  absltest.main()



import pathlib
import sys

from rx.client import grpc_helper
from rx.client import trex_client
from rx.client.commands import command
from rx.client.configuration import config_base
from rx.client.configuration import local


class StopCommand(command.Command):
  """Initialize (or reinitialize) the remote."""

  def run(self) -> int:
    if not self._rxroot:
      sys.stderr.write(f'Could not find rx root in {pathlib.Path.cwd()}\n')
      sys.stderr.flush()
      return -1

    try:
      config = local.create_local_config(self._rxroot)
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.create_authed_client(ch, config)
        client.stop()
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code

    return 0


if __name__ == '__main__':
  print('Call exec.')



import sys
from typing import cast, Optional, Tuple

from absl import logging
from google.protobuf import empty_pb2
import grpc

from rx.client import grpc_helper
from rx.client import login
from rx.client import user
from rx.client import worker_client
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.client.worker import rsync
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

_TIMEOUT = 10

class Client:
  """Handle contacting the remote server."""

  def __init__(
      self,
      channel: grpc.Channel,
      local_cfg: local.LocalConfig,
      auth_metadata: Optional[Tuple[Tuple[str, str]]]):
    self._stub = rx_pb2_grpc.SetupServiceStub(channel)
    self._local_cfg = local_cfg
    self._metadata = local.get_grpc_metadata()
    if auth_metadata:
      self._metadata += auth_metadata

  def create_user_or_log_in(self) -> user.User:
    """This logs in the user, then checks if trex has a username for them."""

    # First, make sure we're logged in with Google.
    lm = login.LoginManager(self._local_cfg.cwd)
    try:
      lm.login()
    except login.AuthError as e:
      raise TrexError(e, -1)
    self._metadata += lm.grpc_metadata
    email = lm.id_token['email']

    # Check if user is set up.
    # TODO: it would be better to read the username from the config and
    # automatically try to set it on the remote. However, this generally
    # shouldn't come up for normal users.
    username = self._get_username(email)
    if not user.has_config(self._local_cfg.cwd):
      with user.CreateUser(self._local_cfg.cwd) as c:
        c['username'] = username
        c['email'] = email
    return user.User(self._local_cfg.cwd, email)

  def dry_run(self):
    rsync.dry_run(self._local_cfg)

  def init(self) -> int:
    try:
      target_env = self._local_cfg.get_target_env()
    except local.ConfigError as e:
      raise TrexError(str(e), -1)
    req = rx_pb2.InitRequest(
      project_name=self._local_cfg['project_name'],
      rsync_source=self._local_cfg.rsync_source,
      target_env=target_env,
    )
    # TODO: create a threaded UserStatus class with __enter__/__exit__.
    sys.stdout.write('Finding a remote worker... ')
    sys.stdout.flush()
    try:
      # Five minute timeout.
      resp = self._stub.Init(req, metadata=self._metadata, timeout=(5 * 60))
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not initialize worker: {e.details()}', -1)
    if resp.result.code != 0:
      raise TrexError(resp.result.message, -1)
    sys.stdout.write('Done.\n')

    with remote.WritableRemote(self._local_cfg.cwd) as r:
      r['workspace_id'] = resp.workspace_id
      r['worker_addr'] = resp.worker_addr
      r['daemon_module'] = resp.rsync_dest.daemon_module

    # Create a container on the worker.
    sys.stdout.write('Setting up the container... ')
    sys.stdout.flush()
    with grpc_helper.get_channel(resp.worker_addr) as ch:
      worker = worker_client.create_authed_client(ch, self._local_cfg)
      worker.init()
    sys.stdout.write('Done.\n')
    print('\nDone setting up rx! To use, run:\n\n\t$ rx <your command>\n')
    return 0

  def stop(self):
    r = remote.Remote(self._local_cfg.cwd)
    req = rx_pb2.StopRequest(workspace_id=r['workspace_id'])
    try:
      result = self._stub.Stop(req, metadata=self._metadata, timeout=_TIMEOUT)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not get user from rx: {e.details()}', -1)

    if result.result.code != 0:
      raise TrexError(
        f'Error stopping machine: {result.result.message}', result.result.code)

  def _create_username(self, email: str) -> str:
    username = user.username_prompt(email)
    req = rx_pb2.SetUsernameRequest(username=username)
    try:
      resp = self._stub.SetUsername(
        req, metadata=self._metadata, timeout=_TIMEOUT)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not set username: {e.details()}', -1)
    if resp.result.code == rx_pb2.INVALID:
      raise TrexError(
        f'{resp.result.message}\nInvalid username: {username}', rx_pb2.INVALID)
    return username

  def _get_username(self, email: str) -> str:
    # Check with rx server.
    username = self._get_username_from_rx()
    if username:
      return username

    # Prompt the user to choose a username.
    return self._create_username(email)

  def _get_username_from_rx(self) -> str:
    try:
      resp = self._stub.GetUser(
        empty_pb2.Empty(), metadata=self._metadata, timeout=_TIMEOUT)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not get user from rx: {e.details()}', -1)
    return resp.username

  def _run_initial_rsync(self, remote_cfg: remote.Remote):
    r = rsync.RsyncClient(self._local_cfg, remote_cfg)
    return_code = r.to_remote()
    if return_code == 0:
      logging.info('Copied files to %s', r.host)


class TrexError(RuntimeError):
  """Class to repackage any init errors that happen and add an exit code."""

  def __init__(self, message, code, *args):
    super().__init__(message, *args)
    self._code = code

  @property
  def code(self):
    return self._code


def create_authed_client(ch: grpc.Channel, local_cfg: local.LocalConfig):
  lm = login.LoginManager(local_cfg.cwd)
  lm.login()
  return Client(ch, local_cfg, lm.grpc_metadata)



import pathlib
import sys

from absl import flags
from absl import logging

from rx.client import grpc_helper
from rx.client import login
from rx.client import menu
from rx.client import trex_client
from rx.client import worker_client
from rx.client.commands import command
from rx.client.configuration import config_base
from rx.client.configuration import local

_DRY_RUN = flags.DEFINE_bool(
  'dry-run', False, 'Shows a list of files that would be uploaded by rx init')


class InitCommand(command.Command):
  """Initialize (or reinitialize) the remote."""

  def __init__(self):
    super().__init__()
    if self._rxroot:
      self._config_exists = local.get_local_config_path(self._rxroot).exists()
      self._user_info_exists = not login.needs_login(self._rxroot)
    else:
      # Fall back on using this dir for rxroot.
      self._rxroot = pathlib.Path.cwd()
      self._config_exists = False
      self._user_info_exists = False

  @property
  def rxroot(self) -> pathlib.Path:
    return self._rxroot

  def _show_init_message(self):
    steps = []
    if self._config_exists:
      reinit = menu.bool_prompt(
        'Looks like you already have an rx workspace. Would you like to stop '
        'that one\nand start a new one?', 'y')
      if not reinit:
        print('Okay, goodbye!')
        sys.exit(0)
      message = 'Got it. To re-init this workspace, rx will:\n\n'
      steps.append('Shut down your existing virtual machine.')
    else:
      message = 'To set up rx, this command will:\n\n'
    if not self._user_info_exists:
      steps.append(
        'Create an account for you with rx (or log you into your existing '
        'account).')
    steps.append('Set up a virtual machine on the cloud (on AWS).')
    steps.append(
      f'Copy over the files in this directory ({self._rxroot}) to your\n'
      '   virtual machine.')
    for num, step in enumerate(steps, 1):
      message += f'{num}. {step}\n'
    if not self._user_info_exists:
      message += (
        '\nWould you like to proceed with logging in/creating an rx account?')
    else:
      message += ('\nWould you like to proceed?')
    if not menu.bool_prompt(message, 'y'):
      print('Okay, goodbye!')
      sys.exit(0)

  def run(self) -> int:
    self._show_init_message()
    if self._config_exists:
      logging.info('Workspace already exists, resetting it.')
    try:
      config = local.create_local_config(self._rxroot)
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.Client(ch, config, auth_metadata=None)
        if _DRY_RUN.value:
          client.dry_run()
          return 0
        if not self._user_info_exists:
          ready_to_login = menu.bool_prompt(
            f"""
Great! First rx will need to open a browser window for you to log in. rx uses
Google's oauth to associate your email with your rx account.

Press y to continue:""", 'y')
          if not ready_to_login:
            print('Okay, goodbye!')
            return 0
        client.create_user_or_log_in()

        if not self._user_info_exists:
          prefix = 'Next,'
        else:
          prefix = 'Great! First'
        ready_to_upload = menu.bool_prompt(
          f"""
{prefix} rx is going to start a virtual machine in the cloud for your user.
This is your private machine and is currently free of charge (though limited in
resources).

The source code in this directory, will be copied to your virtual machine. To
check what will be uploaded, you can rerun this command with --dry-run and
modify .rxignore to exclude anything you don't want copied.

Are you sure you want to upload {self._rxroot} to the cloud?""", 'y')
        if not ready_to_upload:
          print('Okay, goodbye!')
          return 0
        if not menu._QUIET.value:
          print('Great! Let\'s get down to business.')
        return client.init()
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code
    except worker_client.UnreachableError as e:
      print(
        f'Worker {e.worker} was unrechable, run `rx init` to get a new '
        'instance.')
      return e.code

if __name__ == '__main__':
  print('Call exec.')
